* SUPPLY CHAIN MANAGEMENT PROBLEM: DESIGN OF A PETROCHEMICAL SC
* Multiperiod MILP model
* Author: Gonzalo Guillen, Postdoc Carnegie Mellon University

 SCALARS
            INTR interest rate / 0.10 /
            TAX  tax rate / 0.3 /
            SV salvage value / 0.2 /
            DSAT demand satisfaction /0.4/;


 SETS  T time periods / 1*10 /
       I processes / 1*6 /
       J chemicals / 1*18 /
       K markets / 1*10 /
       P plants /1*7/
       W warehouses /1*7/

*Dynamic sets that enable us to define several design problems from a generic data set, each of which includes a specific superstructure and number of time periods

       DINP(P)
       DINW(W)
       DINK(K)
       DINT(T);

*Here I define a supply chain superstructure with 2 plants, 2 warehouses and 4 markets

*DINP('1')=YES;
DINP('5')=YES;
DINP('6')=YES;

*DINW('1')=YES;
DINW('5')=YES;
DINW('6')=YES;

DINK('4')=YES;
DINK('5')=YES;
DINK('7')=YES;
DINK('9')=YES;
*DINK('10')=YES;

*I consider 3 time periods

DINT(T)$(ORD(T) lt 4)=YES;

 ALIAS
(J,JP)
(T,TP)
(P,PP)
(K,KP)
(W,WP)
(I,IP);

 PARAMETERS
             JMM(I) main product for each process  /
                                1  1
                                2  3
                                3  3
                                4  12
                                5  7
                                6 17 / ;

SETS JM(I,*) set that defines the main product for each process;
JM(I,J) = YES $ (ORD(J) EQ JMM(I)) ;

 PARAMETERS

             QELB(P,I,T)         lower bound on plant expansions [tons]
             QEUB(P,I,T)         upper bounds on plant expansions [tons]

             QELBW(W,T)          upper bounds on warehouse expansions [tons]
             QEUBW(W,T)          upper bounds on warehouse expansions [tons]

             MI(I,J)             mass balance coefficients [adimensional]
             ALPHA(P,I,T)        variable investment coefficient [tons]
             BETA(P,I,T)         fixed investment coefficient [$]
             DELTA(P,I,T)        operating expenses per unit of main product [$|ton]
             LAM(P,J,T)          cost of purchasing one unit of chemical j at plant p in period time t [$|ton]
             PUB(P,J,T)          upper bound on chemical j purchased by plant p in period time t [tons]
             PLB(P,J,T)          lower bound on chemical j purchased by plant p in period time t [tons]
             GAMMA(J,K,T)        sales prices [$|ton]
             SUB(J,K,T)          sales upper bound [tons]
             SLB(J,K,T)          sales lower bounds [tons]

             INVINI(J,W)         initial inventory of chemical j kept at warehouse w [tons]
             M1LO(P,W)           lower bound on the material flow between plant p and warehouse w [tons]
             M1UP(P,W)           upper bound on the material flow between plant p and warehouse w [tons]
             M2LO(W,K)           lower bound on the material flow between warehouse w and market k [tons]
             M2UP(W,K)           upper bound on the material flow between warehouse w and market k [tons]
             CAP(J,W)            capacity factor associated with chemical j stored in warehouse w [adimensional]
             TOR(J)              turnover ratio associated with warehouse j [adimensional]

             ALPHAW(W,T)         variable cost associated with warehouse w in period time t [$|ton]
             BETAW(W,T)          investment cost associated with warehouse w in period time t [$]
             DELTAW(J,W,T)       handling cost associated with chemical j stored at warehouse w during period time t [$|ton]
             COSTTRA(P,W,T)      transportation cost from plant p to warehouse w in period time t [$|ton]
             COSTTRB(W,K,T)      transportation cost from warehouse w to market k in period time t [$|ton]
             DISTA(P,W)          distance between plant p and warehouse w [km]
             DISTB(W,K)          distance between warehouse w and market k [km]

             EXCAP(P,I)          existing capacity associated with technology i at plant p [tons]
             EXCAPW(W)           existing capacity of warehouse w [tons] ;


*INPUT DATA: PART OF THE DATA IS RETRIEVED FROM AN EXCEL SPREADSHEET, AND THE REMAINING IS GIVEN IN THIS FILE

*The following parameters are taken from the spreadsheet=======================================================================

$onecho > taskin.txt
par=MI rng=MI!B3:T9 Rdim=1 Cdim=1

par=ALPHA rng=ALPHA!A3:L45 Rdim=2 Cdim=1
par=BETA rng=BETA!A3:L45 Rdim=2 Cdim=1
par=DELTA rng=DELTA!A3:L45 Rdim=2 Cdim=1
par=ALPHAW rng=ALPHAW!A3:K10 Rdim=1 Cdim=1
par=BETAW rng=BETAW!A3:K10 Rdim=1 Cdim=1
par=DELTAW rng=DELTAW!A3:L122 Rdim=2 Cdim=1

par=LAM rng=LAM!A3:L129 Rdim=2 Cdim=1
par=PUB rng=PUB!A3:L129 Rdim=2 Cdim=1
par=GAMMA rng=GAMMA!A3:L183 Rdim=2 Cdim=1
par=SUB rng=SUB!A3:L183 Rdim=2 Cdim=1

par=EN rng=EN!B3:G4 Cdim=1
par=EXCAP rng=EXCAP!A3:G10 Rdim=1 Cdim=1
par=EXCAPW rng=EXCAPW!B3:H4 Cdim=1

par=DISTA rng=DISTA!A3:H10 Rdim=1 Cdim=1
par=DISTB rng=DISTB!A3:K10 Rdim=1 Cdim=1

$offecho

$call gdxxrw.exe SCM-Data_v1.xls @taskin.txt
$gdxin SCM-Data_v1.gdx
$load MI ALPHA BETA DELTA ALPHAW BETAW DELTAW EXCAP EXCAPW LAM PUB GAMMA SUB DISTA DISTB
$gdxin

*The remaining data is next given==================================================================================

*minimum capacity expansions [tons]
 QELB(P,I,T)=10000;
 QEUB(P,I,T)=400000;
 QELBW(W,T)=5000;
 QEUBW(W,T)=400000;

*no lower limits on purchases [tons]
 PLB(P,J,T)=0;

* outsourcing is not permitted (final products cannot be purchased from external suppliers)
loop( (p,j,t,i) $ JM(I,J) ,
 PUB(P,J,T)  = 0 ;
);
* intermediate products cannot be purchased
 PUB(P,'2',T)  = 0 ;
 PUB(P,'7',T)  = 0 ;

* convert ktons into tons
SUB(J,K,T) = SUB(J,K,T) * 1000 ;

* minimum demand constraint (we calculate the minimum sales based on the total demand and the desired minimum demand satisfaction level) [tons]
loop((j,k,t),
SLB(J,K,T) = SUB(J,K,T) * DSAT
);

*Other additional data
* These capacity parameters should take into account the density of the products, since some of them are easier to store. In this example all the products are assumed to be equal.
 CAP(J,W)=1;
*Initial inventory [tons]
 INVINI(J,W)=0;
*Turn-over ratio (adimensional)
 TOR(J)=10;
* Limits on transport flows [tons]
 M1LO(P,W)=5000;
 M1UP(P,W)=500000;
 M2LO(W,K)= 5000;
 M2UP(W,K)=500000;
* To calculate the transportation costs expressed in $ / ton, we multiply $ / kmton (=0.01652) per the distance expressed in km.
* We assume a 5% increase in the cost each year.
loop((p,w,t),
 COSTTRA(P,W,T)= 0.01652 * DISTA(P,W)*(1+0.05*(ord(t)-1));
);
loop((w,k,t),
 COSTTRB(W,K,T)= 0.01652 * DISTB(W,K)*(1+0.05*(ord(t)-1));
);


******************************************************************************
***************************** VARIABLES **************************************
******************************************************************************

 VARIABLES

*continous variables

             NPV                 net present value [$]
             CFLOW(T)            cash flow of period t [$]
             NETE(T)             net earnings [$]
             DEP(T)              depreciation in period t [$]
             FTDC(T)             fraction of total depreciation capital [$]

             Q(P,I,T)            capacity of process i in time period t at plant p [tons]
             QE(P,I,T)           capacity expansions of process i in time period t at plant p [tons]

             QW(W,T)             capacity of warehouse w in time period t [tons]
             QEW(W,T)            capacity expansion of warehouse w in time period t [tons]

             TRA(J,P,W,T)        amount of chemical j transported between plant p and warehouse w in period t [tons]
             TRB(J,W,K,T)        amount of chemical j transported between warehouse w and market k in period t [tons]
             INV(J,W,T)          inventory of chemical j kept at warehouse w in period time t [tons]

             F(P,I,J,T)          flow rates of associated with process i producing chemical j in period t at plant p [tons]
             SA(J,K,T)           sales at market k of chemical j in period t [tons]
             PU(P,J,T)           purchases of chemical j in period t executed by plant p [tons]

*binary variables

             YP(P,I,T)          integer decision variables (1 if technology i at plant p is expanded in period time t otherwise)
             YW(W,T)            integer decision variables (1 if warehouse w is expanded in period time t 0 otherwise)
             X(P,W,T)           integer decision variables (1 if a transportation link is established between p and w in t 0 otherwise)
             Z(W,K,T)           integer decision variables (1 if a transportation link is established between w and k in t 0 otherwise)
             OF                 objective function;

 POSITIVE VARIABLES DEP,FTDC,Q,QE,QW,QEW,TRA,TRB,INV,F,SA,PU;
 BINARY VARIABLES YP,YW,X,Z;

******************************************************************************
***************************** EQUATIONS **************************************
******************************************************************************

 EQUATIONS

             OBJ           net present value

             CASHFLOWS1    calculation of cash flows
             CASHFLOWS2    calculation of cash flows
             NETEARNINGS   calculation of net earnings

             DEPRECIATION(T) depreciation term
             FTDEPCAPITAL(T) fraction of total depreciation capital

             CEUBD         upper bounds on the capacity expansions of the plants
             CELBD         lower bounds on the capacity expansions of the plants

             CAPBL1        capacity balance for period one
             CAPBL         capacity balances considering the expansions beyond period one

             PRCP          production cannot exceed installed capacity
             PBAL          chemical balances (proportionality relations)
             BAL           chemical balances

             TRANSALO      lower bound on transportation limits between plants and warehouses
             TRANSAUP      upper bound on transportation limits between plants and warehouses

             WCEUBD        upper bounds on the capacity expansions of the warehouses
             WCELBD        lower bounds on the capacity expansions of the warehouses
             WCAPBL1       capacity balance for period one
             WCAPBL        capacity balances considering the expansions beyond period one
             WPRCP1        average inventory cannot exceed the installed capacity
             WPRCP2        the inventory cannot exceed the installed capacity
             INV1          material balance in the warehouses for the first time period
             INV2          material balance in the warehouses for periods beyond the first one
             TRANSBLO      lower bound on transportation limits between warehouses and markets
             TRANSBUP      upper bound on transportation limits between warehouses and markets

             PULBD         lower bounds on purchases
             PUUBD         maximum availability of raw materials

             SALBD         lower bounds on sales
             SAUBD         sales cannot exceed the demand
             SATR          the amount transported to the markets equals the demand;

********** Equations: mathematical definition=============================================

 OBJ ..    NPV =E= CFLOW('1') + SUM( T $ ( ORD(T) GT 1 AND DINT(T)) , CFLOW(T) / ((1+INTR)**(ORD(T)-1)));

 CASHFLOWS1(T)$(ORD(T) LT CARD(T))..     CFLOW(T) =E= NETE(T) + DEP(T) - FTDC(T) ;
 CASHFLOWS2(T)$(ORD(T) EQ CARD(T))..     CFLOW(T) =E= NETE(T) + DEP(T) - FTDC(T) + FTDC(T) * CARD(T) * SV ;
 NETEARNINGS(T) $ (DINT(T))..                        NETE(T) =E= (1-TAX) *
                                         (
                                         SUM ( (J,K)$(DINK(K)), GAMMA(J,K,T) * SA(J,K,T) )
                                         - SUM ( (P,J)$(DINP(P)), LAM(P,J,T) * PU(P,J,T) )
                                         - SUM ( (P,I,J)$(DINP(P)), F(P,I,J,T) * DELTA(P,I,T) )
                                         - SUM ( (J,W,K)$(DINW(W) AND DINK(K)), (TRB(J,W,K,T) / TOR(J)) * DELTAW(J,W,T) )
                                         - SUM ( (J,P,W)$(DINP(P) AND DINW(W)), TRA(J,P,W,T) * COSTTRA(P,W,T) )
                                         - SUM ( (J,W,K)$(DINW(W) AND DINK(K)), TRB(J,W,K,T) * COSTTRB(W,K,T) )
                                         - ( ( 1 - SV ) / CARD(T) ) *
                                                 (
                                                 SUM ( (TP,P,I)$(DINP(P) AND DINT(TP)), ALPHA(P,I,TP) * QE(P,I,TP) + BETA(P,I,TP) * YP(P,I,TP) )
                                                 + SUM ( (TP,W)$(DINW(W) AND DINT(TP)), ALPHAW(W,TP) * QEW(W,TP) + BETAW(W,TP) * YW(W,TP) )
                                                 )
                                         );

 DEPRECIATION(T) $ (DINT(T))..               DEP(T) =E= ( ( 1 - SV ) / CARD(T) ) *
                                 (
                                 SUM ( (TP,P,I)$(DINP(P) AND DINT(TP)), ALPHA(P,I,TP) * QE(P,I,TP) + BETA(P,I,TP) * YP(P,I,TP) )
                                 + SUM ( (TP,W)$(DINW(W) AND DINT(TP)), ALPHAW(W,TP) * QEW(W,TP) + BETAW(W,TP) * YW(W,TP) )
                                 ) ;

 FTDEPCAPITAL(T)$ (DINT(T))..               FTDC(T) =E= ( 1 / CARD(T) ) *
                                 (
                                 SUM ( (TP,P,I)$(DINP(P) AND DINT(TP)), ALPHA(P,I,TP) * QE(P,I,TP) + BETA(P,I,TP) * YP(P,I,TP) )
                                 + SUM ( (TP,W)$(DINW(W) AND DINT(TP)), ALPHAW(W,TP) * QEW(W,TP) + BETAW(W,TP) * YW(W,TP) )
                                 ) ;

 CEUBD(P,I,T)$(DINP(P) AND DINT(T)) .. QE(P,I,T) =L= QEUB(P,I,T)*YP(P,I,T);
 CELBD(P,I,T)$(DINP(P) AND DINT(T)) .. QE(P,I,T) =G= QELB(P,I,T)*YP(P,I,T);

 CAPBL1(P,I)$(DINP(P)) ..                  Q(P,I,'1') =E= EXCAP(P,I)+QE(P,I,'1');
 CAPBL(P,I,T) $ (ORD(T) GT 1 AND DINP(P) AND DINT(T)) .. Q(P,I,T) =E= Q(P,I,T-1)+QE(P,I,T);

 PRCP(P,I,T)$(DINP(P) AND DINT(T)) ..                  SUM( J $ JM(I,J), F(P,I,J,T)) =L= Q(P,I,T);

 PBAL(P,I,J,T)$(DINP(P) AND DINT(T)) ..                F(P,I,J,T) =E= ABS(MI(I,J)) * SUM( JP $ JM(I,JP), F(P,I,JP,T) );

 BAL(P,J,T)$(DINP(P) AND DINT(T)) ..                   PU(P,J,T) - SUM(W $(DINW(W)),TRA(J,P,W,T))
                                 =E= SUM ( I, SIGN(MI(I,J)) * F(P,I,J,T) );

 TRANSALO(P,W,T)$(DINP(P) AND DINW(W) AND DINT(T))..      SUM(J,TRA(J,P,W,T)) =G= M1LO(P,W) * X(P,W,T);
 TRANSAUP(P,W,T)$(DINP(P) AND DINW(W) AND DINT(T))..      SUM(J,TRA(J,P,W,T)) =L= M1UP(P,W) * X(P,W,T);

 WCEUBD(W,T)$(DINW(W) AND DINT(T)) ..                  QEW(W,T) =L= QEUBW(W,T)*YW(W,T);
 WCELBD(W,T)$(DINW(W) AND DINT(T)) ..                  QEW(W,T) =G= QELBW(W,T)*YW(W,T);

 WCAPBL1(W)$(DINW(W)) ..                   QW(W,'1') =E= EXCAPW(W)+QEW(W,'1');
 WCAPBL(W,T) $ (ORD(T) NE 1 AND DINW(W) AND DINT(T)) ..  QW(W,T) =E= QW(W,T-1)+QEW(W,T);

 WPRCP1(W,T)$(DINW(W) AND DINT(T)) ..   SUM( (J,K)$(DINK(K)), CAP(J,W) * TRB(J,W,K,T) / TOR(J) ) =L= QW(W,T);
 WPRCP2(W,T)$(DINW(W) AND DINT(T)) ..   SUM(J, INV(J,W,T) * CAP(J,W)) =L= QW(W,T);

 INV1(J,W,T) $ (ORD(T) EQ 1 AND DINW(W) AND DINT(T))..             INVINI(J,W) + SUM(P$(DINP(P)),TRA(J,P,W,T)) =E= INV(J,W,T) + SUM(K$(DINK(K)),TRB(J,W,K,T)) ;
 INV2(J,W,T) $ (ORD(T) GT 1 AND DINW(W) AND DINT(T))..             INV(J,W,T-1) + SUM(P$(DINP(P)),TRA(J,P,W,T)) =E= INV(J,W,T) + SUM(K$(DINK(K)),TRB(J,W,K,T)) ;

 TRANSBLO(W,K,T)$(DINW(W) AND DINK(K) AND DINT(T))..      SUM(J,TRB(J,W,K,T)) =G= M2LO(W,K) * Z(W,K,T);
 TRANSBUP(W,K,T)$(DINW(W) AND DINK(K) AND DINT(T))..      SUM(J,TRB(J,W,K,T)) =L= M2UP(W,K) * Z(W,K,T);

 PULBD(P,J,T)$(DINP(P) AND DINT(T)) .. PU(P,J,T) =G= PLB(P,J,T);
 PUUBD(P,J,T)$(DINP(P) AND DINT(T)) .. PU(P,J,T) =L= PUB(P,J,T);

 SALBD(J,K,T)$(DINK(K) AND DINT(T)).. SA(J,K,T) =G= SLB(J,K,T);
 SAUBD(J,K,T)$(DINK(K) AND DINT(T)).. SA(J,K,T) =L= SUB(J,K,T);
 SATR(J,K,T)$(DINK(K) AND DINT(T))..  SA(J,K,T) =E= SUM(W $ (DINW(W)),TRB(J,W,K,T));

*SOLVE MODEL==========================================================================================

 MODEL SCM /ALL/;
 OPTION MIP = CPLEX;
 OPTION OPTCR = 0;
 SOLVE SCM using MIP maximizing NPV ;

*PRINT RESULTS IN AN EXCEL SPREADSHEET=============================================

FILE SCM_01 /SCM_01.XLS/;
PUT SCM_01;
SCM_01.PC=6;

put 'MODEL',SCM.modelstat,'SOLVER', SCM.solvestat,'Time',SCM.resusd,'NPV',NPV.l;
PUT /;
PUT /;
put 'FIXED PLANTS','FIXED WAREHOUSES','SALES','PURCHASES','OPERATING','HANDLING','TRANSPORT1','TRANSPORT2';
put /;
put (SUM ( (TP,P,I), ALPHA(P,I,TP) * QE.L(P,I,TP) + BETA(P,I,TP) * YP.L(P,I,TP) ) ),
    (SUM ( (TP,W), ALPHAW(W,TP) * QEW.L(W,TP) + BETAW(W,TP) * YW.L(W,TP) ) ),
    (SUM ( (J,K,T), GAMMA(J,K,T) * SA.L(J,K,T) ) ),
    (SUM ( (P,J,T), LAM(P,J,T) * PU.L(P,J,T) ) ),
    (SUM ( (P,I,J,T), F.L(P,I,J,T) * DELTA(P,I,T) ) ),
    (SUM ( (J,W,K,T), (TRB.L(J,W,K,T) / TOR(J)) * DELTAW(J,W,T) ) ),
    (SUM ( (J,P,W,T), TRA.L(J,P,W,T) * COSTTRA(P,W,T) ) ),
    (SUM ( (J,W,K,T), TRB.L(J,W,K,T) * COSTTRB(W,K,T) ) );
put /;
put /;
put 'BINARY(P,I,T)';
put /;
loop(p $ (DINP(P)),
         loop (i,
                 loop(t $ (DINT(T)), put yp.l(p,i,t);
                 );
         put /;
         );
put /;
);
put /;
put /;
put 'Q(P,I,T)';
put /;
loop(p $ (DINP(P)),
         loop (i,
                 loop(t $ (DINT(T)), put q.l(p,i,t);
                 );
         put /;
         );
put /;
);
put /;
put /;
put 'BINARY(W,T)';
put /;
loop(w $ (DINW(W)),
                 loop(t $ (DINT(T)), put yw.l(w,t);
                 );
put /;
);
put /;
put /;
put 'Q(W,T)';
put /;
loop(w $ (DINW(W)),
                 loop(t $ (DINT(T)), put qw.l(w,t);
                 );
put /;
);
put /;
put /;
put 'BINARY(P,W,T)';
put /;
loop(p $ (DINP(P)),
         loop (w $ (DINW(W)),
                 loop(t $ (DINT(T)), put x.l(p,w,t);
                 );
         put /;
         );
put /;
);
put /;
put /;

put 'BINARY(W,K,T)';
put /;
loop(W $ (DINW(W)),
         loop (K $ (DINK(K)),
                 loop(T $ (DINT(T)), put z.l(w,k,t);
                 );
         put /;
         );
put /;
);
put /;
put /;
put 'TRA(J,P,W,T)';
put /;
loop(p $ (DINP(P)),
         loop (w $ (DINW(W)),
                 loop(t $ (DINT(T)), put (SUM( J, TRA.L(J,P,W,T)));
                 );
         put /;
         );
put /;
);
put /;
put /;
put 'TRB(J,W,K,T)';
put /;
loop(W $ (DINW(W)),
         loop (K $ (DINK(K)),
                 loop(T $ (DINT(T)), put (SUM( J, TRB.L(J,W,K,T)));
                 );
         put /;
         );
put /;
);

*END==========================================================================================

